/*
 * Copyright (c) 1996-2000 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <oskit/x86/base_vm.h>		/* phystokv */
#include <oskit/x86/types.h>		/* oskit_addr_t */
#include <oskit/x86/pc/phys_lmm.h>	/* phys_mem_max */
#include <oskit/c/stdio.h>		/* printf */
#include <oskit/c/stdlib.h>		/* atoi */
#include <oskit/c/assert.h>		/* assert */
#include <oskit/c/malloc.h>		/* mustcalloc */
#include <oskit/exec/exec.h>		/* exec_load */

#include "debug.h"

#ifdef NETDISK
#include <oskit/dev/dev.h>              /* All for diskwrite */
#include <oskit/dev/linux.h>
#include <oskit/io/blkio.h>
#include <oskit/fs/read.h>
#include <oskit/c/string.h>
#include <oskit/c/ctype.h>
#include <oskit/diskpart/diskpart.h>

#define DOTIMEIT	/* time the transfer */
#define DOINFLATE	/* support zlib compressed images */
#define SHOWPROGRESS	/* print something once in a while */
#undef DOCHECKSUM	/* compute a checksum of the data read */

#undef DONTREAD		/* testing: define to test write speed of disk */
#undef DONTWRITE	/* testing: define to test read speed of network */

#define MAX_PARTS	30
#define DISKWRITE_SIZE	(16 * 1024)
#define DISKSECT_SIZE	512

#ifdef  DOINFLATE
#include "zlib.h"
#define INFLATEBUF_SIZE	(32 * 1024)
static char inflate_buf[INFLATEBUF_SIZE+DISKSECT_SIZE];
static int iboff;
#endif

#ifdef DOCHECKSUM
unsigned cksum(unsigned char *bp, int len, unsigned cursum);
static unsigned datasum;
#endif

#ifdef DOTIMEIT
#include "timer.h"
typedef unsigned long tstamp_t;
#define GETTSTAMP(ts)	(ts) = currticks()
#if TIMER_FREQ == 100
#define PRINTTSTAMP(ts)	printf("in %ld.%02ld seconds", (ts)/100, (ts)%100)
#else
#define PRINTTSTAMP(ts)	printf("in %d ticks", (ts))
#endif
#else
typedef int tstamp_t;
#define GETTSTAMP(ts)	(ts) = 0
#define PRINTTSTAMP(ts)	
#endif

#endif	/* NETDISK */

#include "netboot.h"			/* NFS_READ_SIZE */
#include "ether.h"			/* ARP_ */
#include "rpc.h"			/* nfs_foo */

#include "boot.h"

#undef min
#define min(a,b)	((a) < (b)? (a) : (b))

static int nfs_port;
static unsigned char dirfh[NFS_FHSIZE];		/* file handle of dir */
static unsigned char filefh[NFS_FHSIZE];	/* file handle of file */

static int kimg_readahead = 16;		/* kbytes to readahead, -1 means max */

/*
 * Callback to read the kernel image from its src, in this case NFS.
 * XXX only handles 32-bit offsets!
 */
static int
kimg_read(void *handle_notused,
	  oskit_addr_t file_ofs, void *buf, oskit_size_t size,
	  oskit_size_t *out_actual)
{
	int n, bytes_read;
	int offset;
	char *bufp = buf;
	int chunk_size;
	int readahead;
#if 0
	printf("kimg_read: file_ofs %x, size %x\n", file_ofs, size);
#endif
	bytes_read = 0;
	offset = file_ofs;
	while (size > 0) {
		chunk_size = min(NFS_READ_SIZE, size);
		if (kimg_readahead < 0)
			readahead = (size / NFS_READ_SIZE) - 1;
		else
			readahead = kimg_readahead;
		n = nfs_read(ARP_ROOTSERVER, nfs_port, filefh,
			     offset, chunk_size, bufp, readahead);
		if (n < 0) {
			printf("Unable to read text: %s\n", nfs_strerror(n));
			return 1;	/* XXX Should be an EX_ constant */
		}
		if (n == 0)
			break;		/* hit eof */
		offset += n;
		bufp += n;
		size -= n;
		bytes_read += n;
	}

        *out_actual = bytes_read;
        return 0;
}

#ifndef NETDISK
/*
 * Callback to load the kernel for real.
 * Copies data from the NFS file into ki->ki_imgaddr.
 */
static int
kimg_load(void *handle,
	  oskit_addr_t file_ofs, oskit_size_t file_size,
	  oskit_addr_t mem_addr, oskit_size_t mem_size,
	  exec_sectype_t section_type)
{
	struct kerninfo *ki = handle;
	oskit_size_t got;
	int err;

#if 0
	printf("** load: file_ofs %x, file_size %x\n", file_ofs, file_size);
	printf("-- load: mem_addr %x, mem_size %x\n", mem_addr, mem_size);
	printf("-- load: sectype = %b\n", section_type, EXEC_SECTYPE_FORMAT);
#endif

        if (!(section_type & EXEC_SECTYPE_ALLOC))
                return 0;

	if ((err = kimg_read(ki, file_ofs,
			     ki->ki_imgaddr + mem_addr - ki->ki_mbhdr.load_addr,
			     file_size, &got)) != 0)
		return err;
	if (got != file_size)
		return EX_CORRUPT;

	return 0;
}

/*
 * Callback for the exec library that just tallys up the sizes.
 */
static int
kimg_tally(void *handle,
	   oskit_addr_t file_ofs, oskit_size_t file_size,
	   oskit_addr_t mem_addr, oskit_size_t mem_size,
	   exec_sectype_t section_type)
{
	struct multiboot_header *h = handle;

#if 0
	printf("** tally: file_ofs %x, file_size %x\n", file_ofs, file_size);
	printf("-- tally: mem_addr %x, mem_size %x\n", mem_addr, mem_size);
	printf("-- tally: sectype = %b\n", section_type, EXEC_SECTYPE_FORMAT);
#endif

	if (!(section_type & EXEC_SECTYPE_ALLOC))
		return 0;

	assert(mem_size > 0);
	if (mem_addr < h->load_addr)
		h->load_addr = mem_addr;
	if (mem_addr+file_size > h->load_end_addr)
		h->load_end_addr = mem_addr+file_size;
	if (mem_addr+mem_size > h->bss_end_addr)
		h->bss_end_addr = mem_addr+mem_size;

	return 0;
}

/*
 * Look in the first MULTIBOOT_SEARCH bytes of the file for the
 * MultiBoot header and puts it in ki->ki_mbhdr;
 *
 * Returns 0 if found, 1 otherwise.
 */
static int
find_multiboot_header(struct kerninfo *ki)
{
	int err;
	oskit_addr_t offset;
	char buf[NFS_READ_SIZE];
	oskit_size_t got;
	int i;
	struct multiboot_header *h;

	for (offset = 0; offset < MULTIBOOT_SEARCH; offset += NFS_READ_SIZE) {
		err = kimg_read(ki, offset, buf, NFS_READ_SIZE, &got);
		if (err)
			return 1;
		for (i = 0; i < got - sizeof(*h); i += 4) {
			h = (struct multiboot_header *)&buf[i];
			if (h->magic == MULTIBOOT_MAGIC
			    && (h->magic + h->flags + h->checksum) == 0) {
				ki->ki_mbhdr = *h;
				return 0;
			}
		}
	}
	return 1;
}

/*
 * Fills in certain fields of the ki_mbhdr multiboot_header using
 * the "executable interpreter" library.
 *
 * Returns 0 on success, 1 otherwise;
 */
static int
fill_in_multiboot_header(struct kerninfo *ki)
{
	struct multiboot_header *h = &ki->ki_mbhdr;
	struct exec_info xinfo;
	int err;

	h->load_addr = 0xffffffff;
	h->load_end_addr = 0;
	h->bss_end_addr = 0;

	err = exec_load(kimg_read, kimg_tally, &ki->ki_mbhdr, &xinfo);
	if (err) {
		printf("Can't tally sections: %s.\n", exec_strerror(err));
		return 1;
	}
	h->entry = xinfo.entry;

	return 0;
}
#endif

/*
 * NFS mount the kernel directory, and perform a lookup on the
 * kernel file.
 *
 * Places the directory filehandle in global "dirfh"
 * Places the kernel filehandle in global filefh
 *
 * returns 0 on success, nonzero on failure
 */
int mountfile_net(struct kerninfo *ki /* IN/OUT */)
{
	int err;
	int mount_port;
	char *readahead_option;

	assert(ki->ki_dirname);
	assert(ki->ki_filename);

	readahead_option = getenv("readahead");
	if (readahead_option)
		kimg_readahead = atoi(readahead_option);
	if (kimg_readahead < 0)
		dprintf("NFS using whole-file readahead\n");
	else if (kimg_readahead > 0)
		dprintf("NFS using %dkb readahead\n", kimg_readahead);

	/*
	 * Lookup NFS/MOUNTD ports for ROOT using PORTMAP.
	 */
	nfs_port = rpc_lookup(ARP_ROOTSERVER, PROG_NFS, 2);
	mount_port = rpc_lookup(ARP_ROOTSERVER, PROG_MOUNT, 1);
	if (nfs_port == -1 || mount_port == -1) {
		printf("Can't get root NFS/MOUNT ports.\n");
		return 1;
	}

	/*
	 * NFS mount the directory.
	 */
	dprintf("Mounting %s...\n", ki->ki_dirname);
	err = nfs_mount(ARP_ROOTSERVER, mount_port, ki->ki_dirname, dirfh);
	if (err) {
		printf("Can't mount ROOT filesystem: %s.\n", nfs_strerror(err));
		return 1;
	}

	/*
	 * Lookup the file.
	 */
	dprintf("Looking up %s...\n", ki->ki_filename);
	err = nfs_lookup(ARP_ROOTSERVER, nfs_port, dirfh, ki->ki_filename,
			 filefh);
	if (err) {
		printf("Can't lookup %s: %s.\n",
		       ki->ki_filename, nfs_strerror(err));
		return 1;
	}

	return 0;
}

#ifndef NETDISK
/*
 * Get the kernel and put it somewhere (not its final position).
 * The kerninfo struct we get passed has the ki_ip, ki_dirname, and ki_filename
 * members filled in.
 * We fill in the ki_mbhdr and ki_imgaddr parts.
 *
 * Returns 0 on success, 1 otherwise.
 */
int
getkernel_net(struct kerninfo *ki /* IN/OUT */)
{
	int err;
	exec_info_t xinfo;

	/*
	 * Mount up
	 */
	if (mountfile_net(ki)) {
		printf("mountfile_net failed.\n");
		return 1;
	}

	/*
	 * Look in the file for the MultiBoot header.
	 */
	if (find_multiboot_header(ki) != 0) {
		printf("Can't find MultiBoot header in image.\n");
		return 1;
	}

	/*
	 * Fill in the rest of the MultiBoot header, if needed.
	 * Stuff like load address, bss size, etc.
	 */
	if (! (ki->ki_mbhdr.flags & MULTIBOOT_AOUT_KLUDGE)
	    && fill_in_multiboot_header(ki) != 0) {
		printf("Can't fill in MultiBoot header.\n");
		return 1;
	}

	/*
	 * Do a bunch of checks on the kernel load ranges.
	 */
	dprintf("Kernel will be at %08x-%08x text+data %d bss %d\n",
		ki->ki_mbhdr.load_addr, ki->ki_mbhdr.bss_end_addr,
		ki->ki_mbhdr.load_end_addr - ki->ki_mbhdr.load_addr,
		ki->ki_mbhdr.bss_end_addr - ki->ki_mbhdr.load_end_addr);
	assert(ki->ki_mbhdr.load_addr < ki->ki_mbhdr.load_end_addr);
	assert(ki->ki_mbhdr.load_end_addr < ki->ki_mbhdr.bss_end_addr);
	if (ki->ki_mbhdr.load_addr < 0x1000) {
		printf("kernel wants to be loaded too low!");
		return 1;
	}
	if (ki->ki_mbhdr.bss_end_addr > phys_mem_max) {
		printf("kernel wants to be loaded beyond available physical memory!");
		return 1;
	}
	if ((ki->ki_mbhdr.load_addr < 0x100000)
	    && (ki->ki_mbhdr.bss_end_addr > 0xa0000)) {
		printf("kernel wants to be loaded on top of I/O space!");
		return 1;
	}

	/*
	 * Now, read the whole thing in somewhere.
	 * It will be copied into its final resting home later,
	 * right before we jump in.
	 */
	printf("%s: loading...", ki->ki_filename);
	ki->ki_imgaddr = mustcalloc(kerninfo_imagesize(ki), 1);
	if ((err = exec_load(kimg_read, kimg_load, (void *)ki, &xinfo)) != 0) {
		printf("Can't load kernel image: %s.\n", exec_strerror(err));
		return 1;
	}

	return 0;			/* yay */
}
#endif

#ifdef NETDISK
char *prog = "netdisk";

void showdiskinfo(char *diskname)
{
	oskit_blkio_t *disk;
	oskit_off_t dsize;
	int dstart, nparts;
	diskpart_t parray[MAX_PARTS], *dpart;
	char dstr[8];
	int err, i, j;

	err = oskit_linux_block_open(diskname,
				     OSKIT_DEV_OPEN_READ|OSKIT_DEV_OPEN_WRITE,
				     &disk);
	if (err)
		return;

	nparts = diskpart_blkio_get_partition(disk, parray, MAX_PARTS);
	if (nparts < 1) {
		dstart = 0;
		err = oskit_blkio_getsize(disk, &dsize);
		if (err) {
			oskit_blkio_release(disk);
			return;
		}
	} else {
		dstart = parray[0].start;
		dsize = parray[0].size;
	}
	snprintf(dstr, sizeof dstr, "%s", diskname);
	printf("Disk %s: sector range: [%d - %d]\n",
	       dstr, dstart, dstart + (int)dsize-1);
	if (nparts < 1) {
		oskit_blkio_release(disk);
		return;
	}

	/*
	 * Note: use BSD syntax
	 */
	dpart = parray[0].subs;
	for (i = 0; i < parray[0].nsubs; i++, dpart++) {
		if (dpart->size == 0)
			continue;
		snprintf(dstr, sizeof dstr, "s%d", i+1);
		printf("  Part %s: [%d - %d]\n",
		       dstr, dpart->start, dpart->start + dpart->size);
		for (j = 0; j < dpart->nsubs; j++) {
			if (dpart->subs[j].size == 0)
				continue;
			snprintf(dstr, sizeof dstr, "s%d%c",
				 i+1, 'a' + j);
			printf("    Part %s: [%d - %d]\n",
			       dstr, dpart->subs[j].start,
			       dpart->subs[j].start + dpart->subs[j].size);
		}
	}
	oskit_blkio_release(disk);
}

int getdiskimage_net(struct kerninfo *ki, char *partition)
{
	static char disk_buf[DISKWRITE_SIZE+DISKSECT_SIZE];
	oskit_off_t filesize, srcoff, dstoff, remaining;
	unsigned long err;
	oskit_size_t got, buffd;
	char *bp;
	char *diskname = 0;
	char *partname = 0;
	char *space;
	oskit_blkio_t *disk = 0;
	oskit_blkio_t *part = 0;
	int nparts;
	diskpart_t part_array[MAX_PARTS], *dpart;
	int retval = 1;  /* Set to 0 for success */
	int minblock;
#ifdef SHOWPROGRESS
	int blockcount = 0;
#endif
#ifdef  DOINFLATE
	z_stream d_stream; /* decompression stream */
	oskit_size_t ibsize = 0;
	int compressed = 0;
#endif
	tstamp_t stamp, estamp;
	struct imagehdr {
		oskit_off_t filesize;
		int magic;
	} *hdr;

	/* Disknames typically start with a letter? */
	diskname = partition;
	if (diskname == 0 || diskname[0] == 0 || !isalpha(*diskname)) {
		printf("Bad disk specification: %s\n",
		       diskname ? diskname : "<null>");
		return 1;
	}

	space = strchr(partition, ' ');
	if (space) {
		*space++ = 0;
	}
	else {
		space = strchr(partition, '\0');
		if (!space) {
			printf("Bad partition specification: %s\n", partition);
			return 1;
		}
	}
	partname = space;
	
	printf("%s: Disk `%s', Partition `%s'\n", prog, diskname, partname);

	/* Create the diskname and partition name from the partition string */

	if (mountfile_net(ki)) {
		printf("mountfile_net failed.\n");
		return 1;
	}

	err = kimg_read(ki, 0, disk_buf, NFS_READ_SIZE, &got);
	if (err != 0) {
		printf("%s: error 0x%lx while reading header.\n", prog, err);
		return 1;
	}
	if (got < NFS_READ_SIZE) {
		printf("%s: short (%d < %d) header read.\n",
		       prog, got, NFS_READ_SIZE);
		return 1;
	}
	hdr = (struct imagehdr *)disk_buf;
	filesize = hdr->filesize;
	srcoff = NFS_READ_SIZE;
	
	if ((unsigned long)(filesize >> 32) != 0) {
		printf("%s: Cannot handle files larger than 4GB yet.\n", prog);
		return 1;
	}

#ifdef  DOINFLATE
	if (got >= sizeof(*hdr) && hdr->magic == 0x69696969) {
		compressed = 1;

		d_stream.zalloc = (alloc_func)0;
		d_stream.zfree = (free_func)0;
		d_stream.opaque = (voidpf)0;
		err = inflateInit(&d_stream);
		if (err != Z_OK) {
			printf("Error initializing inflator\n");
			return 1;
		}
		printf("%s: compressed file size: %lu\n", prog,
		       (unsigned long)filesize);
	} else
#endif
		printf("%s: file size: %lu\n", prog, (unsigned long)filesize);

	if ((long long)filesize < 1) {
		printf("%s: Invalid file size: 0x%8.8lx%8.8lx\n", prog,
		       (unsigned long)(filesize>>32), (unsigned long)filesize);
		return 1;
	}

	err = oskit_linux_block_open(diskname, OSKIT_DEV_OPEN_READ | OSKIT_DEV_OPEN_WRITE, &disk);
	if (err) {
		printf("%s: Disk open failed\n", prog);
		return 1;
	}

	nparts = diskpart_blkio_get_partition(disk, part_array, MAX_PARTS);
	if (*partname != 0 && nparts < 1) {
		printf("%s: No partition table\n", prog);
		retval = 1;
		goto done2;
	}
	if (nparts > 0) {
		if (diskpart_blkio_lookup_bsd_string(part_array, partname,
						     disk, &part) == 0) {
			printf("%s: Couldn't find partition %s\n", prog,
			       partname);
			part = 0;
			goto done2;
		}
	
		dpart = diskpart_lookup_bsd_string(part_array, partname);
		assert(dpart != 0);
		minblock = oskit_blkio_getblocksize(part);

		printf("%s: Partition range: [%d - %d], blocksize: %d\n", prog,
		       dpart->start, dpart->start + dpart->size, minblock);

		if ((oskit_off_t)dpart->size * minblock < filesize) {
			printf("%s: partition %s too small for file\n",
			       prog, partname);
			goto done2;
		}
	} else {
		part = disk;
		minblock = oskit_blkio_getblocksize(part);
	}
	assert(minblock <= DISKSECT_SIZE);

	/* Okay - start reading from NFS and writing to the partition */

	dstoff = 0;

	/* XXX must account for size header; i.e., need the *real* size */
	filesize += srcoff;

	GETTSTAMP(stamp);
	while (srcoff < filesize) {
		/* Buffer up 16k and then write it to the disk... */
		remaining = min(filesize - srcoff, DISKWRITE_SIZE);

#ifndef DONTREAD
		err = kimg_read(ki, (oskit_addr_t)srcoff, disk_buf,
				remaining, &got);
		if (err) {
			printf("%s: NFS read failed, err=0x%lx\n", prog, err);
			retval = 1;
			goto done;
		}
#else
		got = remaining;
#endif
		if (got != remaining) {
			if (got == 0) {
				printf("%s: unexpected EOF\n", prog);
				break;
			}
			printf("%s: NFS short read (%d != %d)\n", prog,
			       got, (int)remaining);
		}

		srcoff += got;
		buffd = got;

#ifdef	DOINFLATE
		/*
		 * The inflate buffer is used to expand the deflated data.
		 *
		 * It may also hold up to ``minblock''-bytes of previously
		 * inflated but not written (residual) data.
		 *
		 * [inflate_buf - inflate_buf+iboff] holds the residual data.
		 * On each pass we attempt to inflate up to INFLATEBUF_SIZE
		 * bytes starting at inflate_buf+iboff.
		 *
		 * After the inflate() call, we write out as many minblock
		 * sized pieces of inflated data as we can, starting at
		 * inflate_buf.  The remainder of the data is moved to the
		 * beginning of the inflate buffer for the next pass.
		 */
		if (compressed) {
			d_stream.next_in   = disk_buf;
			d_stream.avail_in  = buffd;
		inflate_again:
			d_stream.next_out  = inflate_buf + iboff;
			d_stream.avail_out = INFLATEBUF_SIZE;

			err = inflate(&d_stream, Z_SYNC_FLUSH);
			if (err != Z_OK && err != Z_STREAM_END) {
				printf("%s: inflate failed, err=%ld\n",
				       prog, err);
				retval = 1;
				goto done;
			}

			/*
			 * Total data in the inflate buffer is the amount we
			 * converted this time plus any residual we had last
			 * time.
			 */
			ibsize = (INFLATEBUF_SIZE - d_stream.avail_out) +
				iboff;

			/*
			 * Of that, we will write out as many minblock-sized
			 * pieces as possible.  After the write, we will move
			 * the residual up to the top of the inflate buffer
			 * for next time.
			 */
			bp = inflate_buf;
			buffd = ibsize & ~(minblock - 1);
		} else
#endif
			bp = disk_buf;
#ifdef DOCHECKSUM
		datasum = cksum(bp, buffd, datasum);
#endif
#ifndef DONTWRITE
		err = oskit_blkio_write(part, bp, dstoff, buffd, &got);
		if (err) {
			printf("%s: Disk write failed, err=0x%lx\n", prog, err);
			retval = 1;
			goto done;
		}
#else
		got = buffd;
#endif
		if (got != buffd) {
			printf("%s: Disk short write (%d != %d)\n", prog,
			       got, buffd);
			retval = 1;
			goto done;
		}
		dstoff += got;

#ifdef SHOWPROGRESS
		if (++blockcount > 8192) {
			GETTSTAMP(estamp);
			estamp -= stamp;
			printf("Wrote %lu bytes ", (unsigned long)dstoff);
			PRINTTSTAMP(estamp);
			printf("\n");
			blockcount = 0;
		}
#endif
#ifdef  DOINFLATE
		/*
		 * Account for any residual data.
		 * If there is some, we move it to the head of the
		 * inflate buffer.
		 */
		if (compressed) {
			iboff = ibsize - buffd;
			assert(iboff >= 0 && iboff < minblock);
			if (iboff != 0)
				memcpy(inflate_buf, bp+buffd, iboff);
			/*
			 * This will happen if there is not enough room in the
			 * output buffer to decompress the entire input buffer.
			 */
			if (d_stream.avail_in)
				goto inflate_again;
		}
#endif		
	}
	retval = 0;
	printf("\n%s: Success!\n", prog);

 done:
	GETTSTAMP(estamp);
	estamp -= stamp;
	printf("Wrote %lu bytes ", (unsigned long)dstoff);
	PRINTTSTAMP(estamp);
#ifdef DOCHECKSUM
	printf(", checksum=0x%x\n", datasum);
#endif

 done2:
	if (part) oskit_blkio_release(part);
	if (disk) oskit_blkio_release(disk);
	return retval;
}

#ifdef DOCHECKSUM
/*
 * 16-bit checksum, rotating right before each addition;
 * overflow is discarded.
 */
unsigned
cksum(unsigned char *buf, int len, unsigned cursum)
{
	while (len--) {
		if (cursum & 1)
			cursum |= 0x10000;
		cursum = ((cursum >> 1) + *buf++) & 0xffff;
	}
	return cursum;
}
#endif

#endif
