/*
 * Copyright (c) 1995-2001 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of the Flux OSKit.  The OSKit is free software, also known
 * as "open source;" you can redistribute it and/or modify it under the terms
 * of the GNU General Public License (GPL), version 2, as published by the Free
 * Software Foundation (FSF).  To explore alternate licensing terms, contact
 * the University of Utah at csl-dist@cs.utah.edu or +1-801-585-3271.
 * 
 * The OSKit is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with the OSKit; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <malloc.h>

#include <oskit/dev/dev.h>              /* All for diskwrite */
#include <oskit/dev/linux.h>
#include <oskit/io/blkio.h>
#include <oskit/fs/read.h>
#include <oskit/diskpart/diskpart.h>

#include "image.h"
#include "imagehdr.h"
#include "fileops.h"
#include "rpc.h"			/* XXX clear_read_cache() */
#include "perfmon.h"

/* Options */
#define DOINFLATE	/* support zlib compressed images */
#define SHOWPROGRESS	/* print something once in a while */
#undef	DOCHECKSUM	/* compute a checksum of the data read */
#undef  DONTREAD	/* testing: define to test write speed of disk */
#undef  DONTWRITE	/* testing: define to test read speed of network */

/* Input buffer for filedata */
#define MAX_PARTS		30
#define DISKWRITE_SIZE		(32 * 1024)
#define DISKSECT_SIZE		512
static char			disk_buf[DISKWRITE_SIZE+DISKSECT_SIZE];

#ifdef  DOINFLATE
#include "zlib.h"
#define INFLATEBUF_SIZE		(32 * 1024)
static char			*inflate_buf;
static int			iboff;
z_stream			d_stream; /* decompression stream */
oskit_size_t			ibsize    = 0;
oskit_size_t			lastwrite = 0;
int				morestuff = 0;
int				ranout, runahead, notahead, writingdisk;
int				decompress_nextchunk(int compressahead);
#endif

#ifdef DOCHECKSUM
unsigned			cksum(unsigned char *, int, unsigned);
static unsigned			datasum;
#endif

char			*prog = "NetDisk";

int
writeimage(struct in_addr ip, char *dirname, char *filename, char *diskspec)
{
	char		*diskname = 0;
	char		*partname = 0;
	unsigned long	err;
	oskit_size_t	got, buffd;
	oskit_off_t	filesize, srcoff, dstoff, remaining;
	oskit_blkio_t	*disk = 0;
	oskit_blkio_t	*part = 0;
	int		nparts;
	diskpart_t	part_array[MAX_PARTS], *dpart;
	int		retval = 1;  /* Set to 0 for success */
	int		minblock;
	char		*bp;
	struct imagehdr *hdr;
#ifdef SHOWPROGRESS
	int		blockcount = 0;
#endif
#ifdef  DOINFLATE
	int		compressed = 0;
#endif
#ifdef  DOTIMEIT
	tstamp_t	stamp, estamp;
#endif

	/* Disknames typically start with a letter? */
	diskname = diskspec;
	if (diskname == 0 || diskname[0] == 0 || !isalpha(*diskname)) {
		printf("Bad disk specification: %s\n",
		       diskname ? diskname : "<null>");
		return 1;
	}

	/*
	 * The partition is optional but if it exists, disk/part are
	 * separated by a :.
	 */
	partname = diskname;
	strsep(&partname, ":");
	if (!partname)
		partname = "";

	printf("%s: Disk `%s', Partition `%s'\n", prog, diskname, partname);

	if (fileinit(ip, dirname, filename)) {
		printf("%s: file_open failed.\n", prog);
		return 1;
	}

	/*
	 * Read first 1K chunk which has the magic header for a
	 * compressed kernel image.
	 */
	err = fileread(0, disk_buf, MAGICHEADERSIZE, &got);
	if (err != 0) {
		printf("%s: error 0x%lx while reading header.\n", prog, err);
		return 1;
	}
	if (got < MAGICHEADERSIZE) {
		printf("%s: short (%d < %d) header read.\n", prog,
		       got, MAGICHEADERSIZE);
		return 1;
	}
	hdr = (struct imagehdr *)disk_buf;
	filesize = hdr->filesize;
	srcoff = MAGICHEADERSIZE;
	
	if ((unsigned long)(filesize >> 32) != 0) {
		printf("%s: Cannot handle files larger than 4GB yet.\n", prog);
		return 1;
	}

#ifdef  DOINFLATE
	inflate_buf = memalign(1024, INFLATEBUF_SIZE+DISKSECT_SIZE);
	if (!inflate_buf) {
	    printf("%s: Could not allocate inflate buffer\n", prog);
	    return 1;
	}

	if (got >= sizeof(*hdr) && hdr->magic == COMPRESSED_MAGIC) {
		compressed = 1;

		d_stream.zalloc = (alloc_func)0;
		d_stream.zfree = (free_func)0;
		d_stream.opaque = (voidpf)0;
		err = inflateInit(&d_stream);
		if (err != Z_OK) {
			printf("Error initializing inflator\n");
			return 1;
		}
		printf("%s: compressed file size: %lu\n", prog,
		       (unsigned long)filesize);
	} else
#endif
		printf("%s: file size: %lu\n", prog, (unsigned long)filesize);

	if ((long long)filesize < 1) {
		printf("%s: Invalid file size: 0x%8.8lx%8.8lx\n", prog,
		       (unsigned long)(filesize>>32), (unsigned long)filesize);
		return 1;
	}

	err = oskit_linux_block_open(diskname,
				     OSKIT_DEV_OPEN_READ|OSKIT_DEV_OPEN_WRITE,
				     &disk);
	if (err) {
		printf("%s: Disk open failed\n", prog);
		return 1;
	}

	nparts = diskpart_blkio_get_partition(disk, part_array, MAX_PARTS);
	if (*partname != 0 && nparts < 1) {
		printf("%s: No partition table\n", prog);
		retval = 1;
		goto done2;
	}
	if (nparts > 0) {
		if (diskpart_blkio_lookup_bsd_string(part_array, partname,
						     disk, &part) == 0) {
			printf("%s: Couldn't find partition %s\n", prog,
			       partname);
			part = 0;
			goto done2;
		}
	
		dpart = diskpart_lookup_bsd_string(part_array, partname);
		assert(dpart != 0);
		minblock = oskit_blkio_getblocksize(part);

		printf("%s: Partition range: [%d - %d], blocksize: %d\n", prog,
		       dpart->start, dpart->start + dpart->size, minblock);

		if ((oskit_off_t)dpart->size * minblock < filesize) {
			printf("%s: partition %s too small for file\n",
			       prog, partname);
			goto done2;
		}
	} else {
		part = disk;
		minblock = oskit_blkio_getblocksize(part);
	}
	assert(minblock <= DISKSECT_SIZE);

	/* Okay - start reading from NFS and writing to the partition */
	dstoff = 0;

	/* XXX must account for size header; i.e., need the *real* size */
	filesize += srcoff;

	/* Now we start reading full blocks at a time */
	clear_read_cache();

	GETTSTAMP(stamp);
	while (srcoff < filesize) {
		/* Buffer up 16k and then write it to the disk... */
		remaining = min(filesize - srcoff, DISKWRITE_SIZE);

#ifndef DONTREAD
		err = fileread((oskit_addr_t)srcoff,
			       disk_buf, remaining, &got);
		if (err) {
			printf("%s: NFS read failed, err=0x%lx\n", prog, err);
			retval = 1;
			goto done;
		}
#else
		got = remaining;
#endif
		if (got != remaining) {
			if (got == 0) {
				printf("%s: unexpected EOF\n", prog);
				break;
			}
			printf("%s: NFS short read (%d != %d)\n", prog,
			       got, (int)remaining);
		}

		srcoff += got;
		buffd = got;

#ifdef	DOINFLATE
		/*
		 * The inflate buffer is used to expand the deflated data.
		 *
		 * It may also hold up to ``minblock''-bytes of previously
		 * inflated but not written (residual) data.
		 *
		 * [inflate_buf - inflate_buf+iboff] holds the residual data.
		 * On each pass we attempt to inflate up to INFLATEBUF_SIZE
		 * bytes starting at inflate_buf+iboff.
		 *
		 * After the inflate() call, we write out as many minblock
		 * sized pieces of inflated data as we can, starting at
		 * inflate_buf.  The remainder of the data is moved to the
		 * beginning of the inflate buffer for the next pass.
		 */
		if (compressed) {
			d_stream.next_in   = disk_buf;
			d_stream.avail_in  = buffd;
		inflate_again:
			if (decompress_nextchunk(0)) {
				retval = 1;
				goto done;
			}

		write_again:
			/*
			 * We will write out as many minblock-sized
			 * pieces as possible.  After the write, we will move
			 * the residual up to the top of the inflate buffer
			 * for next time.
			 */
			bp = inflate_buf;
			lastwrite = buffd = ibsize & ~(minblock - 1);
			morestuff = 0;
		} else
#endif
			bp = disk_buf;
#ifdef DOCHECKSUM
		datasum = cksum(bp, buffd, datasum);
#endif
#ifndef DONTWRITE
		writingdisk = 1;
		err = oskit_blkio_write(part, bp, dstoff, buffd, &got);
		writingdisk = 0;
		if (err) {
			printf("%s: Disk write failed, err=0x%lx\n", prog, err);
			retval = 1;
			goto done;
		}
#else
		got = buffd;
#endif
		if (got != buffd) {
			printf("%s: Disk short write (%d != %d)\n", prog,
			       got, buffd);
			retval = 1;
			goto done;
		}
		dstoff += got;

#ifdef SHOWPROGRESS
		if (++blockcount > 8192) {
			GETTSTAMP(estamp);
			SUBTSTAMP(estamp, stamp);
			printf("Wrote %qd bytes", dstoff);
			PRINTTSTAMP(estamp);
#ifdef CYCLECOUNTS
			printf(", sleep cycles %qd", total_sleep_cycles);
			printf(", readwait cycles %qd", total_readwait_cycles);
#endif
			printf("\n");
			blockcount = 0;
		}
#endif
#ifdef  DOINFLATE
		/*
		 * Account for any residual data.
		 * If there is some, we move it to the head of the
		 * inflate buffer.
		 */
		if (compressed) {
			/*
			 * This will happen if the upcall decompresses more
			 * stuff while the last diskwrite was waiting to
			 * finish up.
			 */
			if (morestuff)
				goto write_again;
			
			/*
			 * This will happen if there is not enough room in the
			 * output buffer to decompress the entire input buffer.
			 */
			if (d_stream.avail_in)
				goto inflate_again;
		}
#endif		
	}
	retval = 0;
	printf("\n%s: Success!\n", prog);

 done:
	GETTSTAMP(estamp);
	SUBTSTAMP(estamp, stamp);
	printf("Wrote %qd bytes ", dstoff);
	PRINTTSTAMP(estamp);
#ifdef DOCHECKSUM
	printf(", checksum=0x%x\n", datasum);
#endif
	printf("\n");

 done2:
	if (part) oskit_blkio_release(part);
	if (disk) oskit_blkio_release(disk);
	return retval;
}

int
decompress_nextchunk(int compressahead)
{
	unsigned long	err;

	if (compressahead && !writingdisk)
		return 0;

	/*
	 * This ensures we only do one decompress from the upcall.
	 */
	if (morestuff)
		return 0;

	/*
	 * The decompress input buffer is empty, so nothing we can do.
	 */
	if (! d_stream.avail_in) {
		ranout++;
		return 0;
	}

	if (lastwrite) {
		iboff = ibsize - lastwrite;
		assert(iboff >= 0 && iboff < DISKSECT_SIZE);
		if (iboff != 0)
			memcpy(inflate_buf, inflate_buf+lastwrite, iboff);
	}
	else
		iboff = 0;
	
	if (compressahead) {
		runahead++;
	}
	else {
		notahead++;
	}

	d_stream.next_out  = inflate_buf + iboff;
	d_stream.avail_out = INFLATEBUF_SIZE;

	err = inflate(&d_stream, Z_SYNC_FLUSH);
	if (err != Z_OK && err != Z_STREAM_END) {
		printf("%s: inflate failed, err=%ld\n", prog, err);
		return 1;
	}

	/*
	 * Total data in the inflate buffer is the amount we
	 * converted this time plus any residual we had last
	 * time.
	 */
	ibsize = (INFLATEBUF_SIZE - d_stream.avail_out) + iboff;

	morestuff = 1;
	return 0;
}

void
showdiskinfo(char *diskname)
{
	oskit_blkio_t *disk;
	oskit_off_t dsize;
	int dstart, nparts;
	diskpart_t parray[MAX_PARTS], *dpart;
	char dstr[8];
	int err, i, j;

	err = oskit_linux_block_open(diskname,
				     OSKIT_DEV_OPEN_READ|OSKIT_DEV_OPEN_WRITE,
				     &disk);
	if (err)
		return;

	nparts = diskpart_blkio_get_partition(disk, parray, MAX_PARTS);
	if (nparts < 1) {
		dstart = 0;
		err = oskit_blkio_getsize(disk, &dsize);
		if (err) {
			oskit_blkio_release(disk);
			return;
		}
	} else {
		dstart = parray[0].start;
		dsize = parray[0].size;
	}
	snprintf(dstr, sizeof dstr, "%s", diskname);
	printf("Disk %s: sector range: [%d - %d]\n",
	       dstr, dstart, dstart + (int)dsize-1);
	if (nparts < 1) {
		oskit_blkio_release(disk);
		return;
	}

	/*
	 * Note: use BSD syntax
	 */
	dpart = parray[0].subs;
	for (i = 0; i < parray[0].nsubs; i++, dpart++) {
		if (dpart->size == 0)
			continue;
		snprintf(dstr, sizeof dstr, "s%d", i+1);
		printf("  Part %s: [%d - %d]\n",
		       dstr, dpart->start, dpart->start + dpart->size);
		for (j = 0; j < dpart->nsubs; j++) {
			if (dpart->subs[j].size == 0)
				continue;
			snprintf(dstr, sizeof dstr, "s%d%c",
				 i+1, 'a' + j);
			printf("    Part %s: [%d - %d]\n",
			       dstr, dpart->subs[j].start,
			       dpart->subs[j].start + dpart->subs[j].size);
		}
	}
	oskit_blkio_release(disk);
}

void
showdisks(void)
{
	char	buf[32];
	int	i;

	for (i = 0; i < 4; i++) {
		sprintf(buf, "wd%d", i);

		showdiskinfo(buf);
	}
}
